    
    #	by John Walker
    #	http://www.fourmilab.ch/
    
    #	For details, please see: http://www.fourmilab.ch/documents/canon_transits/

    use strict;
    use warnings;

    my $GREGORIAN_EPOCH = 1721425.5;
    

    #   MOD  --  Modulus function which works for non-integers.

    sub mod {
    	my ($a, $b) = @_;
	
	return $a - ($b * floor($a / $b));
    }
    
    #	FLOOR  -- Round number to the nearest integer less than
    #	    	  the argument.  Note that, unlike int(), floor(-1.5) = -2.

    sub floor {
	my $x  = shift;
	my $ix = int($x);
	return (($x >= 0) || ($x == $ix)) ? $ix : ($ix - 1);
    }
    
    #   LEAP_GREGORIAN  --  Is a given year in the Gregorian calendar
    #	    	    	    a leap year ?

    sub leap_gregorian {
    	my $year = shift;
	
	return (($year % 4) == 0) &&
        	(!((($year % 100) == 0) && (($year % 400) != 0)));
    }
    
    #	GREGORIAN_TO_JD  --  Determine Julian day number from Gregorian
    #	    	    	     year, month, day arguments
    
    sub gregorian_to_jd {
	my ($year, $month, $day) = @_;
	
	return ($GREGORIAN_EPOCH - 1) +
               (365 * ($year - 1)) +
               floor(($year - 1) / 4.0) +
               (-floor(($year - 1) / 100.0)) +
               floor(($year - 1) / 400.0) +
               floor((((367 * $month) - 362) / 12) +
               (($month <= 2) ? 0 :
                        	   (leap_gregorian($year) ? -1 : -2)
               ) +
               $day);
    }
    
    #	JD_TO_GREGORIAN  --  Calculate Gregorian calendar date from Julian day,
    #	    	    	     returning an array of year, month, day
    
    sub jd_to_gregorian {
    	my $jd = shift;
	
	my ($wjd, $depoch, $quadricent, $dqc, $cent, $dcent, $quad, $dquad,
            $yindex, $yearday, $leapadj, $year, $month, $day);

	$wjd = floor($jd - 0.5) + 0.5;
	$depoch = $wjd - $GREGORIAN_EPOCH;
	$quadricent = floor($depoch / 146097);
	$dqc = mod($depoch, 146097);
	$cent = floor($dqc / 36524);
	$dcent = mod($dqc, 36524);
	$quad = floor($dcent / 1461);
	$dquad = mod($dcent, 1461);
	$yindex = floor($dquad / 365);
	$year = ($quadricent * 400) + ($cent * 100) + ($quad * 4) + $yindex;
	if (!(($cent == 4) || ($yindex == 4))) {
            $year++;
	}
	$yearday = $wjd - gregorian_to_jd($year, 1, 1);
	$leapadj = (($wjd < gregorian_to_jd($year, 3, 1)) ? 0
                                                      :
                      (leap_gregorian($year) ? 1 : 2)
        	  );
	$month = int(((($yearday + $leapadj) * 12) + 373) / 367);
	$day = ($wjd - gregorian_to_jd($year, $month, 1)) + 1;

	return ($year, $month, $day);
    }
    
    my $JULIAN_EPOCH = 1721423.5;
    
    #	LEAP_JULIAN  --  Determine whether a Julian year number is a leap year

    sub leap_julian {
    	my $year = shift;
	
	return ($year % 4) == (($year > 0) ? 0 : 3);
    }

    #	JULIAN_TO_JD  --  Determine Julian day number from Julian calendar date

    sub julian_to_jd {
    
    	my ($year, $month, $day) = @_;

	#   Adjust negative common era years to the zero-based notation we use.

	if ($year < 1) {
            $year++;
	}

	#   Algorithm as given in Meeus, Astronomical Algorithms, Chapter 7, page 61

	if ($month <= 2) {
            $year--;
            $month += 12;
	}

	return ((floor((365.25 * ($year + 4716))) +
        	floor((30.6001 * ($month + 1))) +
        	$day) - 1524.5);
    }

    #	JD_TO_JULIAN  --  Calculate Julian calendar date from Julian day

    sub jd_to_julian {
    	my $jd = shift;
	my ($z, $a, $b, $c, $d, $e, $year, $month, $day);

	$jd += 0.5;
	$z = floor($jd);

	$a = $z;
	$b = $a + 1524;
	$c = floor(($b - 122.1) / 365.25);
	$d = floor(365.25 * $c);
	$e = floor(($b - $d) / 30.6001);

	$month = int(($e < 14) ? ($e - 1) : ($e - 13));
	$year = int(($month > 2) ? ($c - 4716) : ($c - 4715));
	$day = $b - $d - (int(30.6001 * $e));

	#   If year is less than 1, subtract one to convert from
        #   a zero based date system to the common era system in
        #   which the year -1 (1 B.C.E) is followed by year 1 (1 C.E.).

	if ($year < 1) {
            $year--;
	}

	return ($year, $month, $day);
    }

    #	JD_TO_WEEKDAY  --  Determine day of week from Julian date.
    #	    		   0 = Sunday, 1 = Monday, ... 6 = Saturday.
			   
    our @weekday_names = ( "Sunday", "Monday", "Tuesday", "Wednesday",
                     	   "Thursday", "Friday", "Saturday" );

    sub jd_to_weekday {
    	my $j = shift;
	my $ij = int($j + 1.5);

	$ij %= 7;

	return ($ij < 0) ? (7 + $ij) : $ij;
    }
    
    #	CIVIL_TIME_TO_JD  --  Convert hours, minutes, and seconds to a day fraction
    
    sub civil_time_to_jd {
    	my ($hour, $min, $sec) = @_;
	
	my $s = $sec + (60 * ($min + (60 * $hour)));
	return $s / (24 * 60 * 60);
    }

    #   JD_TO_CIVIL_TIME  --  Convert Julian time to hour, minutes, and seconds.

    sub jd_to_civil_time {
    	my $j = shift;
	my ($ij, $hh, $mm, $ss);

	$j += 0.5;		      # Astronomical to civil
	$ij = int(($j - floor($j)) * 86400.0);
	$hh = int($ij / 3600);
	$mm = int(($ij / 60) % 60);
	$ss = int($ij % 60);
	return ($hh, $mm, $ss);
    }

    #	JD_TO_CALENDAR  --  Convert Julian date to calendar date and time,
    #	    	    	    choosing the calendar based on the Julian
    #	    	    	    day.  If the Julian day is less than
    #	    	    	    $JulianGregorianChange, the Julian calendar
    #	    	    	    is used, if greater than or equal, the
    #	    	    	    Gregorian calendar.
    
    my $JulianGregorianChange = 2299161;    # 1582 October 15 Gregorian
    
    sub setJulianGregorianThreshold {
    	my ($t) = @_;
	
	$JulianGregorianChange = $t;
    }

    sub JDtoCalendar {
    	my ($j) = @_;
	
	my ($calendar, $year, $month, $day);
	
	if (int($j + 0.5) < $JulianGregorianChange) {
	    $calendar = 'J';
	    ($year, $month, $day) = jd_to_julian($j);
	} else {
	    $calendar = 'G';
	    ($year, $month, $day) = jd_to_gregorian($j);
    	}
	my ($hour, $minute, $second) = jd_to_civil_time($j);

	return sprintf("%d-%02d-%02d %02d:%02d %s",
	    $year, $month, $day, $hour, $minute, $calendar);
    }

   #	Built-in test cases
   
if (0) {
    my $jd = 2453249.5;
    my ($y, $m, $d) = jd_to_gregorian($jd);
    my $wd = $weekday_names[jd_to_weekday($jd)];
    
    print("$jd = G $y-$m-$d  $wd\n");
    
    $jd = -2453249.5;
    ($y, $m, $d) = jd_to_gregorian($jd);
    $wd = $weekday_names[jd_to_weekday($jd)];
    print("$jd = G $y-$m-$d $wd\n");
 
  
    $jd = 2453249.5;
    ($y, $m, $d) = jd_to_julian($jd);
    
    print("$jd = J $y-$m-$d\n");
    
    $jd = -2453249.5;
    ($y, $m, $d) = jd_to_julian($jd);
    print("$jd = J $y-$m-$d\n");
}    
    
    1;
